import threading
import base64
import sys
import time
import numpy as np
import matplotlib.pyplot as plt
import cv2
from PIL import Image
from scipy import ndimage
from pylepton import Lepton
from edgetpu.detection.engine import DetectionEngine
from flask import Flask, render_template, Response
from utils import TURBO_COLORMAP, get_temperature, format_temperature


# OpenCV window parameters
WINDOW_NAME = "window"
WINDOW_WIDTH = 400
WINDOW_HEIGHT = 300
# bounding box parameters
LINE_COLOR = (255, 255, 255)
LINE_THICKNESS = 2
# temperature label parameters
LABEL_COLOR = LINE_COLOR
LABEL_FONT = cv2.FONT_HERSHEY_DUPLEX
LABEL_SCALE = 1
LABEL_THICKNESS = 2
# temperature range (in centikelvin)
MIN_TEMP = 29815 # 25 C
MAX_TEMP = 31315 # 40 C
# path to pretrained model
pretrained_model_path = "thermal_face_automl_edge_fast_edgetpu.tflite"

# initialize Flask application
app = Flask(__name__)

# string for encoded data
encoded_data = None


"""
Thermal image capture and processing function
"""
def process():
    print("Started processing thermal image")
    
    # set encoded_img as global variable
    global encoded_data
    
    # initialize OpenCV window
    cv2.namedWindow(WINDOW_NAME, cv2.WINDOW_AUTOSIZE)

    # initialize face detector using pretrained model
    face_detector = DetectionEngine(pretrained_model_path)

    # initialize Lepton communication object
    # start frame capture and processing
    with Lepton() as lepton:
        # store current and previous frame IDs
        curr_frame_id = 1
        prev_frame_id = -1

        while True:
            try:
                # record start time
                start_time = time.time()

                # capture frame from Lepton module
                frame, curr_frame_id = lepton.capture()
                # compare frame IDs
                # skip if they are equal
                if curr_frame_id == prev_frame_id:
                    continue
                # update previous frame ID
                prev_frame_id = curr_frame_id

                # remove last dimension in frame
                frame = np.squeeze(frame, axis=2)
                # upsample
                frame = ndimage.zoom(frame, 5.0)
                # convert centikelvin image to grayscale
                grayscale_img = np.zeros_like(frame)
                grayscale_img[:,:] = frame[:,:]
                grayscale_img = (grayscale_img - MIN_TEMP) / (MAX_TEMP - MIN_TEMP)
                grayscale_img = np.clip(grayscale_img, 0, 1)
                grayscale_img = np.uint8(grayscale_img*255)

                # detect human faces on grayscale image
                faces = face_detector.detect_with_image(
                    Image.fromarray(grayscale_img),
                    threshold=0.5,
                    top_k=10,
                    keep_aspect_ratio=True,
                    relative_coord=False,
                    resample=Image.BILINEAR)

                # apply colormap
                colormap_img = np.zeros_like(grayscale_img)
                colormap_img = TURBO_COLORMAP[grayscale_img]

                # apply bounding boxes
                for face in faces:
                    # extract bounding box coordinates
                    x1 = int(face.bounding_box[0][0])
                    y1 = int(face.bounding_box[0][1])
                    x2 = int(face.bounding_box[1][0])
                    y2 = int(face.bounding_box[1][1])
                    # draw rectangle on colormap image
                    cv2.rectangle(colormap_img, (x1,y1), (x2,y2), LINE_COLOR, LINE_THICKNESS)

                    # get temperature values in the bounding box
                    # uses raw frame from thermal sensor with centikelvin values
                    temperature = get_temperature(frame, x1, y1, x2, y2)

                    if not temperature:
                        continue

                    label = format_temperature(temperature)
                    label_size, _ = cv2.getTextSize(label, LABEL_FONT, LABEL_SCALE, LABEL_THICKNESS)
                    label_position = ((x1+x2)//2 - label_size[0]//2, (y1+y2)//2 - label_size[1]//2)
                    cv2.putText(colormap_img, label, label_position, LABEL_FONT, LABEL_SCALE, LABEL_COLOR, LABEL_THICKNESS, cv2.LINE_AA)

                # calculate timing stats
                duration = time.time() - start_time
                print("Frame took %.f ms (%.2f Hz)" % (duration * 1000, 1 / duration))

                # encode image in JPEG format
                flag, encoded_img = cv2.imencode(".jpg", colormap_img)
                encoded_data = bytearray(encoded_img)

                # display colormap image
                cv2.imshow(WINDOW_NAME, colormap_img)
                cv2.waitKey(1)

            except KeyboardInterrupt:
                break
            
        print("Exit loop")
        
    print("Finished")


@app.route("/")
def index():
    return render_template("index.html")


def gen():
    while True:
        # encode image as base64
        # img = base64.b64encode(encoded_data).decode('utf-8')
        # print(img)
        
        yield (b'--frame\r\n'
               b'Content-Type: image/jpeg\r\n\r\n' + encoded_data + b'\r\n')


@app.route("/stream")
def stream():
    return Response(gen(), mimetype='multipart/x-mixed-replace; boundary=frame')


if __name__ == "__main__":
    # start web server
    # run on a separate thread
    threading.Thread(target=app.run, kwargs=dict(host="0.0.0.0", debug=False)).start()
    
    # start thermal image capture
    # run on a separate thread
    # threading.Thread(target=process).start()
    process()
    
    # app.run(host="0.0.0.0", port=8080, debug=False)